#lang scheme

(require redex/reduction-semantics)
(provide exp?
         tree-exp? tree-stm? 
         pure-exp? lin-stm?
         relop? biop? lvalue? temp? label?  fn?
         bb? tc-exp?
         cub-exp)

(define-language cub-exp
  
  ;; ASTs
  
  (exp (biop exp exp)
       (:= lvalue exp)
       lvalue
       num
       str
       nil
       ()
       (new id exp ...)
       (new-array id exp exp)
       (let (dec ...) exp)
       (begin exp exp exp ...)
       (when exp exp)
       (while exp exp)
       (if exp exp exp)
       (for (id exp exp) exp)
       (break))
  
  (dec (var id exp)
       (var id id exp)
       (type id ty))
  
  (lvalue id
          (dot lvalue num)
          (aref lvalue exp))
  (ty id
      (record id ...)
      (array id))

  ;; language that student type checkers must accept
  
  (tc-exp num
          str
          ()
          id
          (biop tc-exp tc-exp)
          (while tc-exp tc-exp)
          (break)
          (when tc-exp tc-exp)
          (begin tc-exp tc-exp tc-exp ...)
          (:= id tc-exp)
          (if tc-exp tc-exp tc-exp)
          (for (id tc-exp tc-exp) tc-exp)
          (let ((var id tc-exp) ...) tc-exp))
  
  ;; internal tree language
  
  (tree-exp num
            label
            temp
            (biop tree-exp tree-exp)
            (mem tree-exp)
            (call fn tree-exp)
            (eseq tree-stm tree-exp))
  (tree-stm (move (mem tree-exp) tree-exp)
            (move temp tree-exp)
            (texp tree-exp)
            (jump tree-exp label label ...)
            (cjump relop tree-exp tree-exp label label)
            (seq tree-stm tree-stm tree-stm ...)
            label)
  
  ;; internal linear language
  
  (pure-exp num
            label
            temp
            (biop pure-exp pure-exp)
            (mem pure-exp))
  
  (lin-stm mc-stm
           jump-stm
           label)
  
  (jump-stm (jump pure-exp label ...)
            (cjump relop pure-exp pure-exp label label))
  
    
  (mc-stm (move temp pure-exp)
          (move temp (call fn pure-exp))
          (move (mem pure-exp) pure-exp)
          (texp (call fn pure-exp ...)))
  
  ;; basic blocks
  
  (bb (label mc-stm ... jump-stm))

  ;; misc
  
  (biop relop + - * /)
  (relop eqop 
         <= >= < >)
  (eqop = <>)
  
  (label (variable-prefix l:))
  (temp (variable-prefix r:))
  
  (fn "allocate"
      "printstr"
      "printint"
      "printant")
  
  (num natural)
  (str (side-condition (name s string)
                       (andmap (λ (x) (<= 0 (char->integer x) 127))
                               (string->list (term s)))))
  (id (side-condition
       (name 
        var
        (variable-except
         := + - * / and or = <> < <= > >=
         while when for to break let in end function var type array if then else do of nil new))
       (regexp-match #rx"^[a-zA-Z][_a-zA-Z0-9]*$"
                     (symbol->string (term var))))))


(define redex-exp? (redex-match cub-exp exp))
(define redex-dec? (redex-match cub-exp dec))

(define-syntax-rule (pred id? id)
  (define id? 
    (let ([id? (redex-match cub-exp id)])
      (λ (x) 
        (and (id? x) #t)
        #;
        (let ([show? (> (size x) 500)])
          (when show?
            (fprintf (current-error-port) "size! ~a ~s\n" 'id (size x))
            (fprintf (current-error-port) "~s\n" x))
          (begin0 (and (id? x) #t)
                  (when show?
                    (fprintf (current-error-port) "done\n"))))))))
(pred redex-biop? biop)
(pred redex-relop? relop)
(pred lvalue? lvalue)
(pred redex-temp? temp)
(pred redex-label? label)
(pred redex-tree-stm? tree-stm)
(pred redex-tree-exp? tree-exp)
(pred redex-lin-stm? lin-stm)
(pred redex-pure-exp? pure-exp)
(pred redex-tc-exp? tc-exp)
(pred redex-bb? bb)
(pred redex-mc-stm? mc-stm)
(pred jump-stm? jump-stm)
(pred id? id)
(pred fn? fn)

(define (biop? x) (and (symbol? x) (redex-biop? x)))
(define (relop? x) (and (symbol? x) (redex-relop? x)))

(define (label? x)
  (and (symbol? x)
       (redex-label? x)))

(define (temp? x)
  (and (symbol? x)
       (redex-temp? x)))

(define (size x)
  (cond
    [(pair? x) 
     (+ 1 
        (size (car x))
        (size (cdr x)))]
    [else 1]))

(define (tree-exp? x)
  (match x
    [`(,(or '+ '- '* '/) ,e1 ,e2)
     (and (tree-exp? e1) (tree-exp? e2))]
    [`(eseq ,stm ,exp)
     (and (tree-stm? stm)
          (tree-exp? exp))]
    [`(call ,fn ,exp)
     (and (fn? fn)
          (tree-exp? exp))]
    [else
     (and (redex-tree-exp? x) #t)]))

(define (tree-stm? x)
  (match x
    [`(seq ,stms ...)
     (andmap tree-stm? stms)]
    [`(texp ,exp) (tree-exp? exp)]
    [`(move ,(? temp?) ,exp)
     (tree-exp? exp)]
    [else (and (redex-tree-stm? x) #t)]))

(define (pure-exp? x)
  (match x
    [`(,(or '+ '- '* '/) ,e1 ,e2)
     (and (pure-exp? e1) (pure-exp? e2))]
    [else
     (or (number? x)
         (and (redex-pure-exp? x) #t))]))

(define (lin-stm? x)
  (or (mc-stm? x)
      (and (redex-lin-stm? x) #t)))

(define (mc-stm? x)
  (match x
    [`(move ,(? temp?) (call ,(? fn?) ,e))
     (pure-exp? e)]
    [_
     (and (redex-mc-stm? x) #t)]))

(define (tc-exp? x)
  (match x
    [`(,(or '+ '- '* '/ 'while 'when) ,e1 ,e2)
     (and (tc-exp? e1) (tc-exp? e2))]
    [`(let ([var ,ids ,exp1s] ...) ,exp2)
     (and (andmap id? ids)
          (andmap redex-tc-exp? exp1s)
          (tc-exp? exp2))]
    [else
     (and (redex-tc-exp? x) #t)]))

(define (bb? exp)
  (cond
    [(and (list? exp)
          (>= (length exp) 3)
          (and (label? (car exp))
               (let loop ([exp (cdr exp)])
                 (cond
                   [(null? (cdr exp)) 
                    (jump-stm? (car exp))]
                   [else
                    (and (mc-stm? (car exp))
                         (loop (cdr exp)))]))))
     #t]
    [else
     (and (redex-bb? exp) #t)]))

;; adds a few special cases for optimization
(define (exp? x)
  (match x
    [`(,(or '+ '- '* '/ 'while 'when) ,e1 ,e2)
     (and (exp? e1) (exp? e2))]
    [`(let (,decs ...) ,exp)
     (and (andmap redex-dec? decs)
          (exp? exp))]
    [else
     (and (redex-exp? x) #t)]))
